#ifndef __STRUCTUART__
#define __STRUCTUART__

typedef struct UARTdriver {
    int rx_hd ; // reception queue head index
    int rx_tl ; // reception queue tail index
    char rx_q [16]; // reception queue
    int tx_hd ; // transmission queue head index
    int tx_tl ; // transmission queue tail index
    char tx_q [16]; // transmission queue
    int nrx ; // number of characters in rx_queue
    int ntx ; // number of spaces in tx_queue
} UARTdriver;

typedef struct control { // control register fields (uses only ls byte)
    unsigned int ign   : 24,        // ignore uppermost bits
    rts     : 1,         // Request to Send output (bit 7)
    ign2    : 2,         // bits 6,5 ignored
    intTX   : 1,         // interrupt on TX buffer empty (bit 4)
    intRX   : 1,         // interrupt on RX buffer full (bit 3)
    speed   : 3;         // 4,8,16..256 tx-rx clock data rates  (bits 0..2)
} Tcontrol;

typedef struct status { // status register fields (uses only ls byte)
    unsigned int ign : 24,       // ignore uppermost bits
    ign7    : 1,        // ignored (bit 7)
    txEmpty : 1,        // TX register is empty (bit 6)
    rxFull  : 1,        // octet available from RX register (bit 5)
    int_TX_empt: 1,     // interrupt pending on TX empty (bit 4)
    int_RX_full: 1,     // interrupt pending on RX full (bit 3)
    ign2    : 1,        // ignored (bit 2)
    framing : 1,        // framing error (bit 1)
    overun  : 1;        // overun error (bit 0)
}   Tstatus;

typedef union ctlStat { // control + status on same address
    Tcontrol  ctl;        // write-only
    Tstatus   stat;       // read-only
} TctlStat;

typedef union data {    // data registers on same address
    int tx;               // write-only
    int rx;               // read-only
} Tdata;

typedef struct serial {
    TctlStat cs;          // @ (int *)IO_UART_ADDR
    Tdata    d;           // @ (int *)(IO_UART_ADDR+1)
} Tserial;

typedef struct proberx {
    unsigned int
    framing : 1,  // framing error
    overun  : 1,  // overrun
    ign22   : 22, // padding
    rxnum   : 8;  // até 255 caracteres na fila (são <= 16)
} Tproberx;

#endif
