#include "cMIPS.h"
#include "../include/vetorFib.h"
#include "../include/structUart.h"

extern UARTdriver Ud ;
extern Tproberx proberxs ;

volatile Tserial *uart;  // tell GCC not to optimize away code

int proberx () {
    if(proberxs.overun)
        return -1;
    return Ud.nrx;
}

int probetx () {
    return Ud.ntx;
}

int ctrl_util(int speed, int trx, int ttx, int rts) {
    int ctl = 0;
    ctl |= speed;
    ctl |= trx<<3;
    ctl |= ttx<<4;
    ctl |= rts<<6;
    return ctl;
}

void ioctl(int ctl){  
    Tcontrol aux;
    aux.speed = ctl;
    aux.intRX = ctl>>3;
    aux.intTX = ctl>>4;
    aux.rts = ctl>>6;
    uart->cs.ctl= aux;
}

int iostat(){
    int stat = 0;
    stat |= uart->cs.stat.overun;
    stat |= uart->cs.stat.framing<<1;
    stat |= uart->cs.stat.int_RX_full<<3;
    stat |= uart->cs.stat.int_TX_empt<<4;
    stat |= uart->cs.stat.rxFull<<5;
    stat |= uart->cs.stat.txEmpty<<6;
    return stat;
}

int Putc (char c) {
    int ntx = Ud.ntx;
    if (ntx == 0)
        return 0;
    if (ntx == 16 && uart->cs.stat.txEmpty)  // Empty Q && Tx
        uart->d.tx = (unsigned int) c;                   // 32-bits
    else {                                      // Sending smthng
        disableInterr();
        Ud.ntx--;
        // INCREMENTS AND THEN STORE
        Ud.tx_tl = (Ud.tx_tl+1)&0xF;          // modulo 16
        Ud.tx_q[Ud.tx_tl] = c;
        enableInterr();
    }
    return 1;
}

char Getc () {
    char c ;
    if ( Ud.nrx > 0) {
        disableInterr();
        // exclui handler enquanto altera Ud
        Ud.nrx -= 1;
        c = Ud.rx_q [ Ud.rx_hd ];
        Ud.rx_hd = (Ud.rx_hd + 1) & 15; // modulo 16
        enableInterr();
    } else {
        c = -1;
    }
    return c ;
}

int Strlen (char *c) {
    int i= 0;
    while(c[i++]!=10);
    return i-1;
}

int hex2int (int tam, char *hex) {
    int i, ans= 0, exp=1, acc = 0;
    for (i = 1; i < tam; i++)
        exp*=16;
    for (i = 0; hex[i] > 10; i++) {
        acc = 48;
        if (hex[i] > 9+acc)
            acc += 7;
        if (hex[i] > 15+acc) 
            acc += 32;
        if(hex[i] < 0+acc)
            hex[i]=0;
        ans += (hex[i]-acc) * exp;
        exp/=16;
        if(ans > 46)
            return -1;
    }   
    return ans;
}

int int2hex (char *c, int i) {
    int j = 0, val;
    int p = 268435456; //16^7
    if(i < 0){
        c[0] = 'I';
        c[1] = 'N';
        c[2] = 'V';
        c[3] = '\0';
        return 3;
    }
    while (p > 0) {
        val= 0;
        if (i/p) {
            val= i/p;
            i%= p;
        }
        if ((val < 10) && (j || val)) { // impede que '0' sejam escritos à esquerda
            c[j++] = val+'0';
        }
        else if (val > 9)
            c[j++] = 'a'+ (val-10);
        p/=16;
    }   
    c[j] = 0; 
    return j-1;
}

int main(void) { // receive a string through the UART serial interface
    int i, j;    // and responds with the i-eth hex-based fibonacci sequence number
    volatile int state = 0;
    volatile Tstatus status;
    char filona[145], c;
    int hdona = 0;  // head of the big queue
    int tlona = 0;  // tail of the big queue
    int filonx = 0; // store the number of received integers to be processed
    int valfib, qt, rxz, q; // auxiliar variables
    char p[15];
    Tcontrol ctrl;
    uart = (void *)IO_UART_ADDR; // bottom of UART address range

    Ud.ntx = 16; // initializing number of empty spaces in tx queue
    ioctl(ctrl_util(2, 1, 1, 1)); // Speed = 2, intRX = 1, intTX = 1, rts = 1

    do {
        while(!proberx() && !filonx) cmips_delay(1); // wait until first reception

        state++; // state is a volatile variable just for optimize blocking
        if(proberx() == -1){ //overrun checking
            to_stdout('O');
            to_stdout('V');
            to_stdout('E');
            to_stdout('R');
            to_stdout('U');
            to_stdout('N');
            to_stdout('\0');
            return state;
        }
        while((c=Getc()) != -1){ // dump the reception queue to the big queue
            filona[tlona] = c;
            tlona++;
            tlona%=145;
            if(c=='\n'){
                filonx++;
                filona[tlona] = 10;
            }
        }

        while(filonx > 0){
            qt = Strlen(filona+hdona);
            valfib = hex2int(qt, (filona+hdona));
            if(valfib != -1)          // if is a valid value (lower than 2c and bigger or equal to 0)
                valfib = buf[valfib]; // if not valid, UART receives and prints ffffffff
            hdona += qt+1;
            hdona %= 145;
            qt = int2hex(p, valfib);
            j = -1;
            do {
                j++;
                while((rxz = proberx())<9 && (rxz >= 0) && !(q=Putc(p[j]))) cmips_delay(1);

                if(rxz == -1){ // overrun checking
                    to_stdout('O');
                    to_stdout('V');
                    to_stdout('E');
                    to_stdout('R');
                    to_stdout('U');
                    to_stdout('N');
                    to_stdout('\0');
                    return state;
                }

                while((c=Getc()) != -1){
                    filona[tlona] = c;
                    tlona++;
                    tlona%=145;
                    if(c=='\n'){
                        filonx++;
                        filona[tlona] = 10;
                    }
                }

                if(!q)                 
                    j--;
                
            } while (p[j] != 0);
            filonx--;

            if(!filonx){
                return state;
            }
        }
    } while (1);
    return state;

}
